@@ -9,15 +9,16 @@ use Log::ger;
 # put other modules alphabetically here
 use IO::Interactive qw(is_interactive);
 use List::Util qw(first);
-use Mo qw(build default);
+use Moo;
+
 #use Moo;
 extends 'Perinci::CmdLine::Base';
 
 # put global variables alphabetically here
 our $AUTHORITY = 'cpan:PERLANCAR'; # AUTHORITY
-our $DATE = '2021-10-02'; # DATE
+our $DATE = '2021-12-01'; # DATE
 our $DIST = 'Perinci-CmdLine-Lite'; # DIST
-our $VERSION = '1.911'; # VERSION
+our $VERSION = '1.912'; # VERSION
 
 # when debugging, use this instead of the above because Mo doesn't give clear
 # error message if base class has errors.
@@ -31,9 +32,6 @@ has validate_args => (
     is=>'rw',
     default => 1,
 );
-has plugins => (
-    is => 'rw',
-);
 
 my $formats = [qw/text text-simple text-pretty json json-pretty csv termtable html html+datatables perl/];
 
@@ -108,10 +106,6 @@ sub BUILD {
     $self->{formats} //= $formats;
 
     $self->{per_arg_json} //= 1;
-
-    Perinci::CmdLine::Base::__plugin_activate_plugins_in_env();
-    Perinci::CmdLine::Base::__plugin_activate_plugins(@{ $self->{plugins} })
-          if $self->{plugins};
 }
 
 my $setup_progress;
@@ -154,36 +148,6 @@ sub hook_after_parse_argv {
     }
 }
 
-# BEGIN_BLOCK: equal2
-sub equal2 {
-    state $require = do { require Scalar::Util };
-
-    # for lack of a better name, currently i name this 'equal2'
-    my ($val1, $val2) = @_;
-
-    # here are the rules:
-    # - if both are undef, 1
-    # - undef equals nothing else
-    # - if both are ref, equal if their refaddr() are equal
-    # - if only one is ref, 0
-    # - if none is ref, compare using eq
-
-    if (defined $val1) {
-        return 0 unless defined $val2;
-        if (ref $val1) {
-            return 0 unless ref $val2;
-            return Scalar::Util::refaddr($val1) eq Scalar::Util::refaddr($val2);
-        } else {
-            return 0 if ref $val2;
-            return $val1 eq $val2;
-        }
-    } else {
-        return 0 if defined $val2;
-        return 1;
-    }
-}
-# END_BLOCK: equal2
-
 sub hook_before_parse_argv {
     my ($self, $r) = @_;
 
@@ -260,7 +224,7 @@ sub hook_before_action {
         # it
         last if $meta->{features} && $meta->{features}{validate_vars};
 
-        Perinci::CmdLine::Base::__plugin_run_event(
+        $self->_plugin_run_event(
             name => 'validate_args',
             r => $r,
             on_success => sub {
@@ -283,7 +247,7 @@ sub hook_before_action {
                     #say "D:we have pre-compiled validator codes";
 
                     for my $arg (sort keys %{ $meta->{args} // {} }) {
-                        next unless exists($r->{args}{$arg});
+                        next unless exists($r->{args}{$arg}) || $meta->{args}{$arg}{'x.perinci.cmdline.default_from_schema'};
 
                         # we don't support validation of input stream because
                         # this must be done after each 'get item' (but periswrap
@@ -301,7 +265,7 @@ sub hook_before_action {
                     my %validators_by_schema; # key = "$schema"
                     require Data::Sah;
                     for my $arg (sort keys %{ $meta->{args} // {} }) {
-                        next unless exists($r->{args}{$arg});
+                        next unless exists($r->{args}{$arg}) || $meta->{args}{$arg}{'x.perinci.cmdline.default_from_schema'};
 
                         # we don't support validation of input stream because
                         # this must be done after each 'get item' (but periswrap
@@ -325,15 +289,23 @@ sub hook_before_action {
               DO_VALIDATE: {
                     for my $arg (sort keys %{ $meta->{args} // {} }) {
                         my $v = $validators_by_arg{$arg} or next;
+
+                        # we want to get default value from schema, but do not
+                        # want to make unspecified args spring into existence
+                        # with 'undef' values. so we record the existence first
+                        # here.
+                        my $arg_exists = exists $r->{args}{$arg};
+
                         my $res = $v->($r->{args}{$arg});
                         if ($res->[0]) {
                             die [400, "Argument '$arg' fails validation: $res->[0]"];
                         }
-                        my $val0 = $r->{args}{$arg};
-                        my $coerced_val = $res->[1];
-                        $r->{args}{$arg} = $coerced_val;
-                        $r->{args}{"-orig_$arg"} = $val0
-                            unless equal2($val0, $coerced_val);
+                        if ($arg_exists || $meta->{args}{$arg}{'x.perinci.cmdline.default_from_schema'}) {
+                            my $val0 = $r->{args}{$arg};
+                            my $coerced_val = $res->[1];
+                            $r->{args}{$arg} = $coerced_val;
+                            $r->{args}{"-orig_$arg"} = $val0;
+                        }
                     }
                 } # DO_VALIDATE
 
@@ -663,7 +635,7 @@ Perinci::CmdLine::Lite - A Rinci/Riap-based command-line application framework
 
 =head1 VERSION
 
-This document describes version 1.911 of Perinci::CmdLine::Lite (from Perl distribution Perinci-CmdLine-Lite), released on 2021-10-02.
+This document describes version 1.912 of Perinci::CmdLine::Lite (from Perl distribution Perinci-CmdLine-Lite), released on 2021-12-01.
 
 =head1 SYNOPSIS
 
@@ -878,21 +850,17 @@ perlancar <perlancar@cpan.org>
 
 =head1 CONTRIBUTORS
 
-=for stopwords Paul Cochrane Steven Haryanto (on Asus Zenbook)
+=for stopwords Paul Cochrane Steven Haryanto
 
 =over 4
 
 =item *
 
-Paul Cochrane <paul.cochrane@posteo.de>
-
-=item *
-
 Paul Cochrane <paul@liekut.de>
 
 =item *
 
-Steven Haryanto (on Asus Zenbook) <stevenharyanto@gmail.com>
+Steven Haryanto <stevenharyanto@gmail.com>
 
 =back
 
